#!/usr/bin/env python3
"""Run black in a docker"""

import logging
import os
import sys
from argparse import ArgumentParser
from typing import List, Optional

from .config import ODOO_11, ODOO_12, ODOO_13, ODOO_14, ODOO_15, ODOO_16, Config
from .docker_build import add_build_options, build_local_image, get_build_options
from .docker_client import DockerClient, get_volumes
from .parsing import apply, basic_parser

__version__ = "2.3.0"
__date__ = "2022-03-03"
__updated__ = "2024-04-17"

_logger = logging.getLogger(__name__)
_BLACK_DEST = "black"
"""Name of destination variable used for black in parsing"""


def __parser() -> ArgumentParser:
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_short_description}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2022 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)

    parser.add_argument(
        help="Run black on given directory (relative path, for example "
        "github.com/OCA/server-auth)",
        dest=_BLACK_DEST,
    )
    parser.add_argument(
        "--write",
        help="Write changes",
        action="store_true",
    )

    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    add_build_options(parser)
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    return black(getattr(nmspc, _BLACK_DEST), nmspc.write, get_build_options(nmspc))


def uses_black(odoo_type: str) -> bool:
    """Return true if black is used for the modules of the odoo type"""
    return odoo_type in (ODOO_11, ODOO_12, ODOO_13, ODOO_14, ODOO_15, ODOO_16)


def black(directory: str, write: bool, build_options, show_version: bool = True):
    """Run black"""
    config = Config()
    if not uses_black(config.odoo_type):
        raise Exception(f"No black for {config.odoo_type}")

    build_local_image(build_options)

    pwd = os.environ["PWD"]
    volumes, path = get_volumes(pwd, directory, write)

    command = ["black", "."]
    # The line was in a configuration file but not in modules in that Odoo 11 modules
    # so this is still needed to avoid reformatting when forgetting to update the module
    # configuration.
    if config.odoo_type == ODOO_11:
        command.extend(("--line-length", "79"))
    if not write:
        command.extend(("--check", "--diff"))
    command = [" ".join(command)]
    if show_version:
        command.insert(0, "black --version")

    _logger.info("Running black in %s", path)
    return DockerClient.run(
        config.local_image,
        config.local_tag,
        {
            "entrypoint": "sh",
            "user": os.getuid(),
            "command": ["-c", ";".join(command)],
            "volumes": volumes,
            "workdir": path,
        },
        False,
    )


if __name__ == "__main__":
    sys.exit(main())
