#!/usr/bin/env python3
"""Script to locally build a docker image"""

import argparse
import datetime
import logging
import os
import signal
import sys
from subprocess import call, check_output, run
from typing import List, Optional

import hglib  # type: ignore[import]
from python_on_whales import docker

from .config import ODOO_7, ODOO_8, ODOO_10, ODOO_11, ODOO_13, ODOO_14, Config
from .docker_build_clean import clean
from .docker_build_copy import Configuration as CopyConfiguration
from .docker_build_copy import add_build_copy_options, copy, get_build_copy_options
from .docker_client import capture_docker_exit_code
from .parsing import apply, basic_parser
from .storage import get_orus_api_token

_logger = logging.getLogger(__name__)

__version__ = "1.7.0"
__date__ = "2018-04-04"
__updated__ = "2024-01-17"


def add_build_options(parser: argparse.ArgumentParser):
    """Add build options to parser"""
    parser.add_argument(
        "--build-arg",
        help="build arg for the image, formatted like FOO=BAR [default: %(default)s]",
        default=None,
        nargs="*",
    )
    add_build_copy_options(parser)


# XXX Using something else than redone arguments would be better
def get_build_options(namespace: argparse.Namespace) -> List[str]:
    """Return build options usable in :func:`build_local_image`."""
    result = []
    if namespace.build_arg:
        for arg in namespace.build_arg:
            result.append("--build-arg")
            result.append(arg)
    result += get_build_copy_options(namespace)
    return result


def __parser() -> argparse.ArgumentParser:
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_short_description}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2018-2024 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    add_build_options(parser)
    # TODO add tag option, and maybe force tag option
    parser.add_argument(
        "--ensureconf",
        help="ensureconf [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--push", help="Push image [default: %(default)s]", action="store_true"
    )
    parser.add_argument(
        "--dev",
        help="add dev feature to generated image [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--no-pull",
        help="indicate to docker to not pull the base image [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--local",
        help="Use local image name",
        action="store_true",
        dest="local",
        default=None,
    )
    # TODO (maybe) add argument for other build arg
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Parse arguments and docker build"""
    # TODO the script assume it is launched from the parent super project
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)
    # TODO changez pour une vérification de la présence d’un Dockerfile
    if project_name == "odoo_scripts":
        _logger.fatal("You must run this script from the super project")
        return 1

    # TODO add a way to store configuration options in a project file
    # Argument parsing
    parser = __parser()

    nmspc = parser.parse_args(argv)
    apply(nmspc)
    ensureconf = nmspc.ensureconf
    dev = nmspc.dev
    push = nmspc.push

    dockerfile = "Dockerfile"
    if not os.path.exists(dockerfile):
        _logger.critical("No dockerfile found")
        return 1

    c = Config()
    odoo_type = c.odoo_type
    repository = c.local_image_name if nmspc.local else c.repository
    _logger.debug("Docker repository: %s", repository)
    # TODO ensureconf
    if ensureconf:
        raise NotImplementedError
    # call build copy
    copy(**CopyConfiguration(nmspc))

    # clean on exit
    def signal_handler(code, frame):
        # TODO handle the case when odoo_scripts is not installed
        clean()
        # XXX needed?

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # used to retag
    tags: List[str] = []

    buildargs = {}
    if os.path.exists(".hg"):
        mercurial = hglib.open(project_path)
        tags = (
            mercurial.rawcommand(
                hglib.client.cmdbuilder(b"identify", template=b"{tags}")
            )
            .decode("UTF-8")
            .split(":")
        )
        # remove pseudo-tag
        if "tip" in tags:
            tags.remove("tip")
        buildargs["VERSION"] = tags[0] if tags else "unreleased"
        description = mercurial.rawcommand(
            hglib.client.cmdbuilder(
                b"log",
                r=b".",
                template=b"{latesttag}{sub('^-0-.*', '', '-{latesttagdistance}-h{node|"
                b"short}')}",
            )
        ).decode("UTF-8")
        buildargs["VERSION_ARG"] = description
        # SENTRY_RELEASE is only used in old Dockerfiles
        buildargs["SENTRY_RELEASE"] = description
        buildargs["VCS_URL"] = mercurial.paths(b"default").decode("UTF-8")
        buildargs["VCS_REF"] = mercurial.identify(id=True).decode("UTF-8")
    elif os.path.exists(".git"):
        description = check_output(["git", "describe"]).decode()
        # SENTRY_RELEASE is only used in old Dockerfiles
        buildargs["SENTRY_RELEASE"] = description
        complete_process = run(["git", "describe", "--tags", "--exact-match"])
        if complete_process.returncode == 0:
            tags_string = complete_process.stdout.decode()
        if tags_string:
            tags = tags_string.split()
        buildargs["VERSION"] = tags[0] if tags else "unreleased"
        git_sha = check_output(["git", "rev-parse", "HEAD"]).decode()
        buildargs["VERSION_ARG"] = tags[0] if tags else git_sha
    buildargs["BUILD_DATE"] = datetime.datetime.now().isoformat()
    if nmspc.build_arg:
        for arg in nmspc.build_arg:
            a = arg.split("=")
            buildargs[a[0]] = a[1]
    _logger.debug("Build args: %s", buildargs)
    if nmspc.orus_api_token or get_orus_api_token():
        buildargs["PIP_INDEX_URL"] = (
            f"https://__token__:{nmspc.orus_api_token or get_orus_api_token()}@orus.io/api/v4/groups/9/-/packages/pypi/simple"  # noqa: E501
        )
    else:
        if c.orus_api_token_required:
            _logger.error(
                "No Orus API token provided, the project indicates it is mandatory"
            )
    if dev:
        debug_dockerfile = "Dockerfile.debug"
        call(["cp", dockerfile, debug_dockerfile])

        with open(debug_dockerfile, "a") as myfile:
            myfile.write("\n# Developer helpers\nENV PIP_CACHE_DIR=/var/cache/pip\n")
            if odoo_type in (
                ODOO_7,
                ODOO_8,
                ODOO_10,
                ODOO_11,
                ODOO_13,
                ODOO_14,
            ):
                myfile.write("USER root\nRUN apt-get update -qq\n")
                myfile.write(
                    "RUN --mount=target=/var/cache/apt,type=cache,sharing=shared "
                    "DEBIAN_FRONTEND=noninteractive apt-get install -y -qq "
                )
            if odoo_type in (ODOO_11, ODOO_13, ODOO_14):
                myfile.write(
                    "python3-watchdog python3-ipdb python3-pyinotify python3-wheel\n"
                )
                # newer setuptools_scm>6 is not compatible with python 3.5 due
                # to the use of f-strings
                myfile.write(
                    "RUN --mount=target=/var/cache/pip,type=cache,sharing=shared "
                    "python3 -m pip install "
                    "\"setuptools_scm<6;python_version<='3.5'\"\n"
                )
                # Install pdb++
                myfile.write(
                    "RUN --mount=target=/var/cache/pip,type=cache,sharing=shared "
                    "python3 -m pip install pdbpp\n"
                )
                myfile.write("USER odoo\n")
            elif odoo_type in (ODOO_10,):
                myfile.write("python-ipdb\n")
                myfile.write(
                    "RUN --mount=target=/var/cache/pip,type=cache,sharing=shared pip "
                    "install watchdog --disable-pip-version-check "
                    "--system --only-binary wheel"
                )
            elif odoo_type in (ODOO_7, ODOO_8):
                myfile.write("python-ipdb\n")
                myfile.write(
                    "RUN --mount=target=/var/cache/pip,type=cache,sharing=shared pip "
                    "install pyinotify\n"
                )
            else:
                myfile.write(
                    "RUN --mount=target=/var/cache/pip,type=cache,sharing=shared "
                    "python3 -m pip install pdbpp watchdog ipdb pyinotify\n"
                )

        dockerfile = debug_dockerfile
        # TODO remove temp image

    pull = not nmspc.no_pull
    _logger.debug("Docker Pull %s", pull)
    full_tags = [repository] if nmspc.local else [f"{repository}:{tag}" for tag in tags]
    with capture_docker_exit_code() as exit_code:
        if docker.buildx.is_installed():
            _logger.info("Using docker buildx")
            docker.buildx.build(
                context_path=".",
                pull=pull,
                build_args=buildargs,
                tags=full_tags,
                file=dockerfile,
            )
        else:
            _logger.info("Using docker build (quiet)")
            previous_value = os.environ.get("DOCKER_BUILDKIT", None)
            os.environ["DOCKER_BUILDKIT"] = "1"
            docker.image.legacy_build(
                context_path=".",
                pull=pull,
                build_args=buildargs,
                tags=full_tags,
                file=dockerfile,
            )
            if previous_value is None:
                os.environ.pop("DOCKER_BUILDKIT")
            else:
                os.environ["DOCKER_BUILDKIT"] = previous_value
            _logger.info("Built %s", ",".join(full_tags))

    if dev:
        call(["rm", dockerfile])

    if push and not exit_code:
        _logger.info("Docker push %s", repository)
        docker.push(full_tags)
    # XXX call cleanup more intelligently
    signal_handler(0, None)
    return int(exit_code)


def build_local_image(build_options: List[str], force: bool = False) -> int:
    """Build the local image if necessary
    :param force: force build if True
    :return: 0 if success, docker failed error code otherwise
    """
    c = Config()
    image = c.local_image_name

    image_list = docker.image.list(image)
    if not image_list:
        _logger.info("Image %s does not exist", image)
    else:
        _logger.info("Image %s exists", image)
    if not image_list or force:
        _logger.info("Building image %s", image)
        arguments = ["--dev", "--local", *build_options]
        failed = main(arguments)
        if failed:
            return failed
    return 0


if __name__ == "__main__":
    sys.exit(main())
