#!/usr/bin/env python3
"""Run isort in a docker"""

import argparse
import logging
import os
import sys
from typing import List, Optional

from .config import (
    ODOO_7,
    ODOO_8,
    ODOO_9,
    ODOO_10,
    ODOO_11,
    ODOO_12,
    ODOO_13,
    ODOO_14,
    ODOO_15,
    ODOO_16,
    Config,
)
from .docker_build import add_build_options, build_local_image, get_build_options
from .docker_client import DockerClient, get_volumes
from .parsing import apply, basic_parser

__version__ = "3.1.0"
__date__ = "2020-09-10"
__updated__ = "2024-04-17"

_logger = logging.getLogger(__name__)
_ISORT_DEST = "isort"
"""Name of destination variable used for isort in parsing"""
_ISORT_DIR_DEST = "isort_dir"

CHECK_MODE = "check"
"""Run isort to check"""
DIFF_MODE = "diff"
"""Run isort to show differences"""
WRITE_MODE = "write"
"""Run isort to write differences"""


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2020, 2022-2024 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_isort_group(parser: argparse.ArgumentParser):
    """Add isort option (--isort/--no-isort) to the given parser.
    Defaults to not using isort."""
    isort_group = parser.add_mutually_exclusive_group()
    isort_group.add_argument(
        "--isort-check",
        help="Run isort (check only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=CHECK_MODE,
    )
    isort_group.add_argument(
        "--isort-diff",
        help="Run isort (check only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=DIFF_MODE,
    )
    isort_group.add_argument(
        "--no-isort",
        help="Do not run isort [default]",
        action="store_const",
        dest=_ISORT_DEST,
        const=None,
    )


def parser_add_isort(parser: argparse.ArgumentParser):
    """Add isort option (--isort/--no-isort) to the given parser.
    Defaults to not using isort."""
    parser.add_argument(
        help="Run pylint on given directory (relative path, for example "
        "github.com/OCA/server-auth)",
        dest=_ISORT_DIR_DEST,
    )
    isort_group = parser.add_mutually_exclusive_group()
    isort_group.add_argument(
        "--check",
        help="Run isort (check only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=CHECK_MODE,
        default=CHECK_MODE,
    )
    isort_group.add_argument(
        "--diff",
        help="Run isort (diff only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=DIFF_MODE,
    )
    isort_group.add_argument(
        "--write",
        help="Run isort (write)",
        action="store_const",
        dest=_ISORT_DEST,
        const=WRITE_MODE,
    )


def apply_isort(namespace: argparse.Namespace, config: Optional[Config] = None):
    """Run isort if the option was set."""
    if _ISORT_DEST in namespace and getattr(namespace, _ISORT_DEST):
        if config is None:
            config = Config()
        if uses_isort(config.odoo_type):
            return isort(
                getattr(namespace, _ISORT_DEST),
                directory=getattr(namespace, _ISORT_DIR_DEST),
                config=config,
            )
        else:
            _logger.warning(f"No isort for {config.odoo_type}, ignoring running it")
    return 0


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    add_build_options(parser)
    parser_add_isort(parser)
    nmspc = parser.parse_args(argv)
    config = Config()
    apply(nmspc)
    if config.odoo_type not in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        build_local_image(get_build_options(nmspc))
    return apply_isort(nmspc, config=config)


def uses_isort(odoo_type: str) -> bool:
    """Return true if isort is used for the modules of the odoo type"""
    return odoo_type in (
        ODOO_7,
        ODOO_8,
        ODOO_9,
        ODOO_10,
        ODOO_11,
        ODOO_12,
        ODOO_13,
        ODOO_14,
        ODOO_15,
        ODOO_16,
    )


def isort(mode: str, directory: str = "", config: Optional[Config] = None):
    """Run isort"""
    if config is None:
        config = Config()
    odoo_type = config.odoo_type
    if not uses_isort(odoo_type):
        raise Exception(f"No isort for {odoo_type}")

    command = []

    if mode == CHECK_MODE:
        command.append("--check")
    if mode == DIFF_MODE:
        command.append("--check")
        command.append("--diff")

    # determine image to use based on odoo version
    if odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        repository = "xcgd/isort"
        tag = "odoo"
        pull = True
    else:
        repository = config.local_image
        tag = config.local_tag
        pull = False

        if odoo_type not in (ODOO_11, ODOO_12, ODOO_13):
            # append the directory, needed with isort≥5.0.0
            command.append(".")

    pwd = os.environ["PWD"]
    volumes, path = get_volumes(pwd, directory, mode == WRITE_MODE)
    _logger.info("Running isort in %s", path)
    return DockerClient.run(
        repository,
        tag,
        {
            "entrypoint": "isort",
            "user": os.getuid(),
            "command": command,
            "volumes": volumes,
            "workdir": path,
        },
        pull,
    )


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
