#!/usr/bin/env python3
"""Run a postgresql docker with volume based on project name."""

import argparse
import atexit
import logging
import os
import sys
import tempfile
import time
from functools import partial
from typing import Callable, List, Optional, Tuple

from python_on_whales import Container, docker
from python_on_whales.exceptions import NoSuchContainer

from .config import Config
from .docker_client import DockerClient, Mount, mount
from .parsing import apply, basic_parser

_logger = logging.getLogger(__name__)

__version__ = "1.1.0"
__date__ = "2020-11-09"
__updated__ = "2023-10-31"

POSTGRES_PASSWORD = "this-could-be-anything"
PSQL = "psql"
"""Command to use to run psql in the docker container."""
SH = "SH"
"""Command to use to run sh in the docker container."""


def docker_run_postgresql(
    project_name: str,
    postgresql_version: str,
    host_pg_port: Optional[int] = None,
    stop_at_exit: bool = True,
) -> Tuple[Container, Callable, Optional[str]]:
    """

    :param project_name:
    :param postgresql_version:
    :param host_pg_port: if None, put the socket in a temp directory,
     otherwise publish the port
    :param stop_at_exit: True to stop the container at exit
    :return: container, method to call to stop it, socket directory
    """
    pg_repository = "postgres"
    version = "{}-alpine".format(postgresql_version)
    pg_image = "{}:{}".format(pg_repository, version)
    name = "postgresql-{}-{}".format(postgresql_version, project_name)
    # need to pull otherwise the containers.create might fail
    docker.image.pull(f"{pg_repository}:{version}")
    pg_data_volume_name = "postgresql_{}-{}".format(postgresql_version, project_name)
    path: Optional[str] = None
    env = {"POSTGRES_PASSWORD": POSTGRES_PASSWORD}

    def stop_postgresql(pg_container: Container):
        # TODO test if still exists
        #  maybe by looking at pg.status
        # TODO factorized with other methods that start dockers
        try:
            _logger.info("Stopping postgresql %s", pg_container.name)
            pg_container.stop()
            docker.wait(pg_container)
            _logger.info("Removing postgresql container %s", pg_container.name)
            pg_container.remove()
        except NoSuchContainer:
            _logger.info("Postgresql container %s already stopped", pg_container.name)
        # crash because docker or pg changes the owner of the directory
        # temp_directory.cleanup()

    for container in docker.container.list():
        if name == container.name:
            _logger.info("Postgresql Container already running")
            for mount_ in container.mounts:
                if mount_.destination == "/var/run/postgresql":
                    path = str(mount_.source)
            return container, partial(stop_postgresql, container), path

    mounts: List[Mount] = [
        mount(
            "/var/lib/postgresql/data",
            pg_data_volume_name,
            "volume",
        )
    ]

    port_bindings: List[
        Tuple[str | int, str | int] | Tuple[str | int, str | int, str]
    ] = []
    if isinstance(host_pg_port, int):
        port_bindings.append((5432, host_pg_port))
    else:
        path = tempfile.TemporaryDirectory(prefix="odoo_scripts_postgres_socket-").name
        os.makedirs(path)
        mounts.append(mount("/var/run/postgresql", path, "bind"))
    _logger.debug("Creating postgresql container")
    pg = docker.container.create(
        pg_image,
        mounts=mounts,
        publish=port_bindings,
        name=name,
        envs=env,
    )

    _logger.info("Starting postgresql container %s", pg.name)
    pg.start()

    stop_pg = partial(stop_postgresql, pg)

    if stop_at_exit:
        atexit.register(stop_pg)
    _logger.debug("Waiting for postgres to start")
    # give pg the time to start up
    time.sleep(5)
    return pg, stop_pg, path


def __parser() -> argparse.ArgumentParser:
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2020 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)
    parser.add_argument(
        "--name",
        help="Container name to use [default: %(default)s]",
        default=project_name,
    )
    parser.add_argument(
        "-c",
        "--command",
        help="Command to run. If specified the command will not be interactive"
        ", otherwise it will be interactive.",
        metavar="command",
        default=None,
    )
    exec_group = parser.add_mutually_exclusive_group()
    exec_group.add_argument(
        "--psql",
        help="Exec psql (default)",
        action="store_const",
        const=PSQL,
        dest="exec",
        default=PSQL,
    )
    exec_group.add_argument(
        "--sh",
        help="Exec a shell",
        action="store_const",
        const=SH,
        dest="exec",
    )
    # no simple way to indicate that these options are only used with psql exec
    parser.add_argument("-d", "--database", help="Database name", default=None)
    parser.add_argument("-u", "--user", help="User", default=None)
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Run postgresql docker, and exec a psql into it"""
    parser = __parser()
    namespace = parser.parse_args(argv)
    apply(namespace)
    project_name = namespace.name
    exec_type = namespace.exec
    command = namespace.command
    user = namespace.user
    # get postgres version from config
    c = Config()
    postgresql_version = c.postgresql_version
    pg, stop_pg, path = docker_run_postgresql(project_name, postgresql_version)
    # run an exec command
    docker_command = None
    interactive = not bool(command)
    if exec_type == PSQL:
        if not user:
            user = "postgres"
        docker_command = ["psql", "--user", user]
        if command:
            _logger.info("Exec command in the container")
            docker_command.extend(["-c", command])
        else:
            _logger.info("Exec psql in the container")
        if namespace.database:
            docker_command.append(namespace.database)
        user = "root"
    elif exec_type == SH:
        if not user:
            user = "root"
        docker_command = ["sh"]
        if command:
            _logger.info("Exec shell command in the container")
            docker_command.extend(["-c", command])
        else:
            _logger.info("Exec sh in the container")
    if docker_command:
        return DockerClient.exec(pg, docker_command, user, interactive)
    return 0


if __name__ == "__main__":
    sys.exit(main())
