#!/usr/bin/env python3
"""Run a rednerd docker with volume based on project name.

Documentation: http://127.0.0.1:12345/docs if the server is started with port 12345
"""

import argparse
import logging
import os
import sys
import tempfile
import time
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import quote

import requests
import requests_unixsocket  # type: ignore[import]
from psycopg2 import OperationalError
from python_on_whales import docker

from .config import Config
from .docker_client import DockerClient, DockerService, Mount, mount
from .parsing import apply, basic_parser
from .postgres import host_from_dsn, postgres_connect_args, postgres_connect_parser

_logger = logging.getLogger(__name__)

__version__ = "1.4.0"
__date__ = "2021-01-14"
__updated__ = "2023-12-06"

MIGRATE = "migrate"
ADMIN_PASSWORD = "admin-password"
SERVE = "serve"


class DockerRednerd(DockerService):
    """
    This class allow to start the rednerd server
    """

    def __init__(
        self,
        project_name: str,
        repository: str,
        db_user: str,
        db_password: Optional[str],
        database: str,
        db_host: Optional[str] = None,
        db_port: Optional[str] = None,
        attach: bool = True,
        stop_at_exit: bool = True,
        redner_user: Optional[str] = None,
        redner_password: Optional[str] = None,
        redner_admin_password: Optional[str] = None,
        redner_port: Optional[int] = None,
        use_socket: bool = False,
        version: str = "latest",
        force_pull: bool = False,
        log_level: str = "info",
    ):
        """
        :param project_name: name of the project, used for some volume and container
        :param repository: repository to use for the redner image
        :param db_user: redner database user
        :param db_password: redner database password
        :param database: database name for rednerd, the database must already be created
        :param db_host: redner database host (unix socket or hostname)
        :param db_port: database port
        :param redner_port: Port to use for http, if None, no port is given to rednerd
        :param use_socket: Set to true to make rednerd use a Unix socket
        :param force_pull: Indicate to pull the latest version of the specified image,
            otherwise the image will only be pulled if not present
        """
        assert project_name is not None, "Missing project_name"
        assert repository is not None, "Missing repository"
        assert database is not None, "Missing database"

        # db_password can be none, in that case expect that any password would work
        super().__init__(
            f"rednerd_{project_name}",
            repository,
            version,
            False,
            attach=attach,
            stop_at_exit=stop_at_exit,
        )
        self.redner_user = redner_user
        self.redner_password = redner_password
        self.redner_admin_password = redner_admin_password
        self.db_host: Optional[str]
        dsn = f"dbname={database}"
        if db_user:
            dsn += f" user={db_user}"
        if db_password:
            dsn += f" password={db_password}"
        if db_port:
            dsn += f" port={db_port}"
        if db_host:
            dsn += f" host={db_host}"
            self.db_host = db_host
        else:
            try:
                self.db_host = host_from_dsn(dsn)
            except OperationalError:
                _logger.info("Unable to connect with dsn %s", dsn)
                self.db_host = None
        self.db_is_socket_path = self.db_host.startswith("/") if self.db_host else False
        self.environment = {
            "REDNERD_ENVIRONMENT": "local",
            "REDNERD_LOG_FORMAT": "pretty",
            "REDNERD_DSN": dsn,
            "REDNERD_LOG_LEVEL": log_level,
            # errors when not defined when using standard docker
            "NO_CONFD": "true",
        }
        if redner_port is not None:
            self.redner_port = str(redner_port)
        # create a socket directory
        if use_socket:
            self.socket_dir = tempfile.TemporaryDirectory(
                prefix="odoo_scripts_rednerd_socket-"
            ).name
            """Rednerd socket’s directory."""
            os.mkdir(self.socket_dir)
            self.socket_path = os.path.join(self.socket_dir, "rednerd.sock")
        self.force_pull = force_pull
        self.api_key = None
        """Store the API key"""

    def _start(self) -> None:
        """Start rednerd container"""
        # base parameters for container creation
        ports: List[Tuple[str, str]] = []
        mounts: List[Mount] = []
        create_param: Dict[str, Any] = dict(
            name=self.name,
            publish=ports,
            envs=self.environment,
            mounts=mounts,
            networks="host",
        )
        if self.db_is_socket_path:
            if self.db_host:
                mounts.append(mount(self.db_host, self.db_host, "bind"))
            else:
                _logger.error(
                    "Connection to db uses a socket but missing path, not mounting"
                )
        # migrate
        _logger.info("migrate")
        DockerClient.run(
            repository=self.repository,
            tag=self.tag,
            run_kwargs=dict(create_param, command=[MIGRATE]),
        )
        # setup admin password if provided
        if self.redner_admin_password is not None:
            _logger.info("admin-password")
            DockerClient.run(
                repository=self.repository,
                tag=self.tag,
                run_kwargs=dict(
                    create_param, command=[ADMIN_PASSWORD, self.redner_admin_password]
                ),
            )
        # serve
        create_param["image"] = self.base_image
        command = [SERVE, "--token-secret", self.name]
        schemes = []
        if hasattr(self, "redner_port"):
            command.extend(["--port", str(self.redner_port)])
            schemes.append("http")
        if hasattr(self, "socket_path"):
            command.extend(["--socket-path", self.socket_path])
            schemes.append("unix")
            _logger.info("Using socket %s", self.socket_path)
        if schemes:
            command.extend(["--scheme", ",".join(schemes)])
        create_param["command"] = command
        if hasattr(self, "redner_port"):
            # useless when using --network=host
            ports.append((f"{self.redner_port}", self.redner_port))
            _logger.info("Using port %s", self.redner_port)
        if hasattr(self, "socket_dir"):
            mounts.append(mount(self.socket_dir, self.socket_dir, "bind"))
        self._container = docker.container.create(**create_param)
        _logger.info("Starting container %s", self.name)
        _logger.debug(create_param)
        self._container.start()
        session: Optional[requests.sessions.Session] = None
        base_url: Optional[str] = None
        if hasattr(self, "socket_path"):
            session = requests_unixsocket.Session()
            base_url = "http+unix://{}".format(quote(self.socket_path, safe=""))
        elif hasattr(self, "redner_port"):
            session = requests.sessions.Session()
            base_url = f"http://localhost:{self.redner_port}"
        # Start log process immediately to get any log now
        if not self.attach:
            self._start_log_process()
        # if there is neither port or socket, no session nor base_url
        # wait for redner to be started
        if hasattr(self, "socket_path"):
            _logger.info("Waiting for redner socket")
            while not os.path.exists(self.socket_path):
                time.sleep(0.1)
        elif hasattr(self, "redner_port") and session and base_url:
            _logger.info("Waiting for port to be available")
            ready = False
            while not ready:
                try:
                    session.get(base_url)
                    ready = True
                except requests.exceptions.ConnectionError:
                    pass
                time.sleep(0.2)
        else:
            _logger.warning("No port nor socket to wait for")
        # change socket permissions, current version has too much restrictive values
        # it might be due to requests_unixdomain that want more than available
        if hasattr(self, "socket_path"):
            _logger.info("Changing socket permission")
            DockerClient.exec(
                self._container, ["chmod", "-R", "777", self.socket_path], "root"
            )

        if (
            base_url is not None
            and session is not None
            and self.redner_admin_password is not None
        ):
            _logger.info("Log in as admin")
            # auth
            request = session.post(
                f"{base_url}/api/v1/auth/authenticate",
                json={
                    "username": "admin",
                    "password": self.redner_admin_password,
                    "set-cookie": True,
                },
            )
            request.raise_for_status()
            cookie_jar = request.cookies
            # check if user exists
            request = session.get(
                f"{base_url}/api/v1/user/",
                cookies=cookie_jar,
            )
            request.raise_for_status()
            user_missing = True
            for user in request.json():
                if user["username"] == self.redner_user:
                    user_missing = False
                    continue
            if user_missing:
                _logger.info("Creating user %s", self.redner_user)
                request = session.post(
                    f"{base_url}/api/v1/user",
                    json={"username": self.redner_user, "is-admin": False},
                    cookies=cookie_jar,
                )
                request.raise_for_status()
            # change user password
            _logger.info("Creating %s password", self.redner_user)
            request = session.post(
                f"{base_url}/api/v1/user/{self.redner_user}/password",
                json={"password": self.redner_password},
                cookies=cookie_jar,
            )
            request.raise_for_status()
            _logger.info("Grabbing API key for %s", self.redner_user)
            request = session.get(
                f"{base_url}/api/v1/user/{self.redner_user}/apikey",
                json={},
                cookies=cookie_jar,
            )
            request.raise_for_status()
            api_keys_list = request.json()
            name = "docker_rednerd"
            if len(api_keys_list) > 0:
                # delete our api key to retrieve its value again after creating it
                for api_keys_data in api_keys_list:
                    if api_keys_data["name"] == name:
                        request = session.delete(
                            f"{base_url}/api/v1/user/{self.redner_user}/apikey/{name}",
                            json={},
                            cookies=cookie_jar,
                        )
                        request.raise_for_status()
            _logger.info("Creating new API key")
            request = session.post(
                f"{base_url}/api/v1/user/{self.redner_user}/apikey",
                json={
                    "name": name,
                    "description": "Odoo scripts created API KEY",
                },
                cookies=cookie_jar,
            )
            self.api_key = request.json()["value"]


def get_default_redner_database(project_name: str) -> str:
    return f"redner_{project_name}"


def __parser() -> argparse.ArgumentParser:
    """Return a parser for docker_rednerd"""
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".")[0]
    program_license = f"""{program_short_description}

      Created by Oury Balde on {__date__}.
      Copyright 2021, 2023 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)

    parser = basic_parser(program_license, program_version_message)
    postgres_connect_parser(
        parser,
        "db",
        default_user="redner",
        default_password="redner",
        default_database=get_default_redner_database(project_name),
    )
    parser.add_argument(
        "--name",
        help="Base container name to use [default: %(default)s]",
        default=project_name,
    )
    parser.add_argument(
        "--user",
        help="Redner user to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--password",
        help="Redner password to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--admin-password",
        help="Redner admin password to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Redner port to use, use a socket if empty [default: %(default)s]",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--use-socket",
        help="Specify Redner socket to use [default: %(default)s]",
        action="store_true",
        default=False,
    )
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Run redner docker using command line arguments"""
    parser = __parser()
    namespace = parser.parse_args(argv)
    apply(namespace)
    project_name = namespace.name
    connect_args = postgres_connect_args(namespace, "db")
    user = namespace.user
    password = namespace.password
    admin_password = namespace.admin_password
    port = namespace.port
    use_socket = namespace.use_socket

    # get rednerd version and repository from config
    config = Config()
    rednerd_version = config.rednerd_version
    rednerd_repository = config.rednerd_repository

    if rednerd_repository is None:
        _logger.fatal("Missing repository name for rednerd image")
        return 81
    # Instantiate docker rednerd
    docker_rednerd = DockerRednerd(
        project_name,
        rednerd_repository,
        db_user=connect_args["user"],
        db_password=connect_args["password"],
        database=connect_args["database"],
        db_host=connect_args["host"],
        db_port=connect_args["port"],
        attach=True,
        redner_user=user,
        redner_password=password,
        redner_admin_password=admin_password,
        redner_port=port,
        version=rednerd_version,
        use_socket=use_socket,
    )
    docker_rednerd.start()

    return 0


if __name__ == "__main__":
    sys.exit(main())
