#!/usr/bin/env python3
"""Imports CSV files into an Odoo server via Odoo import wizard."""

import hashlib
import hmac
import logging
import sys
from datetime import datetime
from typing import Dict, List, Optional

from requests_toolbelt import MultipartEncoder  # type: ignore[import]

from . import odoo_connect_parser, odoo_login
from .importing import add_importing_file_parsing, extract_model_lang_from_parsed
from .parsing import apply

_logger = logging.getLogger(__name__)

__version__ = "1.0.5"
__date__ = "2019-07-29"
__updated__ = "2022-01-20"


def import_with_base_import(
    login: str,
    password: str,
    database: str,
    protocol: str,
    host: str,
    port: int,
    timeout: int,
    model_filenames: List[Dict[str, str]],
    update_parameter: bool = True,
) -> int:
    odoo, session_id = odoo_login(
        database, host, login, password, port, protocol, timeout
    )
    if not odoo.version.startswith("11.0"):
        _logger.fatal("Base Import is not compatible with version %s", odoo.version)
        return 2

    import_obj = odoo.env["base_import.import"]
    param_obj = odoo.env["ir.config_parameter"]

    _logger.info("Getting a CSRF token")
    # Reproduce what odoo.http::csrf_token does to build a CSRF token.
    max_ts = ""
    db_secret = param_obj.get_param("database.secret")
    csrf_token = "%so%s" % (
        hmac.new(
            db_secret.encode(),
            msg=("%s%s" % (session_id, max_ts)).encode(),
            digestmod=hashlib.sha1,
        ).hexdigest(),
        max_ts,
    )
    _logger.debug("CSRF token: %s", csrf_token)

    _logger.info("Importing CSV files")

    import_options = {
        "encoding": "utf-8",
        "quoting": '"',
        "separator": ",",
        "headers": True,
    }
    _logger.info("%d files to import", len(model_filenames))
    for element in model_filenames:
        model = element["model"]
        csv_file = element["filename"]
        lang = element["lang"]
        _logger.info("Importing - %s in model %s (lang %s)", csv_file, model, lang)
        context = {
            "lang": lang,
            # TODO add tz maybe
        }
        import_dlg_id = import_obj.create({"res_model": model}, context=context)
        with open(csv_file, "rb") as file:
            multipart = MultipartEncoder(
                fields={
                    "csrf_token": csrf_token,
                    "import_id": str(import_dlg_id),
                    "file": ("filename", file, "text/csv"),
                }
            )
            data = multipart.to_string()
            response = odoo.http(
                "base_import/set_file",
                data=data,
                headers={"Content-Type": multipart.content_type},
            )
            # in Odoo, context contains user lang, but that is ok so far
            _logger.debug("response: %s", response)
        # This returns a lot of stuff we don't need, except for the field list.
        fields = import_obj.parse_preview(import_dlg_id, import_options)["headers"]
        # Got fields; now run the import.
        errors = []
        # this is where setting the context is necessary
        results = import_obj.do(import_dlg_id, fields, import_options, context=context)
        for result in results:
            if result["type"] == "error":
                errors.append(result["message"])
        if errors:
            _logger.fatal("Error while importing - %s in model %s:", csv_file, model)
            for error in errors:
                _logger.warning(" - %s", error)
            return 1
        _logger.info("Imported - %s in model %s", csv_file, model)

    if update_parameter:
        _logger.debug("Setting import_csv parameter")
        param_obj.set_param("import_base_import.write_date", str(datetime.now()))

    _logger.info("Done.")
    return 0


def main(argv: Optional[List[str]] = None) -> int:
    """Parse arguments and launch conversion"""
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

  Created by Vincent Hatakeyama on %s.
  Copyright 2019, 2020, 2023 XCG Consulting. All rights reserved.

  Licensed under the MIT License

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
""" % (
        program_shortdesc,
        str(__date__),
    )
    # Argument parsing
    parser = odoo_connect_parser(program_license, program_version_message)
    add_importing_file_parsing(parser)

    nmspc = parser.parse_args(argv)
    apply(nmspc)

    return import_with_base_import(
        login=nmspc.login,
        password=nmspc.password,
        port=nmspc.port,
        host=nmspc.host,
        protocol=nmspc.protocol,
        timeout=nmspc.timeout,
        database=nmspc.database,
        model_filenames=extract_model_lang_from_parsed(nmspc),
    )


if __name__ == "__main__":
    sys.exit(main())
