"""Function to help create and parse options."""

import argparse
import logging
from typing import Optional

import coloredlogs  # type: ignore[import]

from . import __version__


def add_verbosity_to_parser(parser: argparse.ArgumentParser):
    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="count",
        help="Increase verbosity level [default: %(default)s]",
        default=0,
    )
    verbosity_group.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="count",
        help="Decrease verbosity level [default: %(default)s]",
        default=0,
    )


def logging_from_verbose(namespace: argparse.Namespace):
    verbosity = namespace.verbose - namespace.quiet
    if verbosity == 0:
        level = logging.INFO
    elif verbosity == -1:
        level = logging.WARNING
    elif verbosity == -2:
        level = logging.ERROR
    elif verbosity == -3:
        level = logging.CRITICAL
    elif verbosity > 0:
        level = logging.DEBUG
    coloredlogs.install(level, fmt="%(asctime)s %(levelname)8s %(name)s %(message)s")


def basic_parser(description: Optional[str], version: str) -> argparse.ArgumentParser:
    """Create a basic parser"""
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="{} [odoo_scripts {}]".format(version, __version__),
    )
    add_verbosity_to_parser(parser)

    return parser


def apply(namespace: argparse.Namespace):
    """Use the basic parser options"""
    logging_from_verbose(namespace)
